"""Module for sending requests to BEA API and handling errors"""

import requests


class BEAAPI(object):
    r"""Base class for communicating with the API, outputs raw JSON results.

    This class communicates with the BEA API and handles possible errors,
    returning unfiltered JSON results in case of success. It stores your API key
    on initiation, and uses it for future requests. It serves as a base class to
    enable the :class:`.BEA` class to abstract from requests and error handling.

    Parameters
    ----------
    api_key
        Your BEA API key
    """

    # Base BEA API url
    _base_url = r'http://apps.bea.gov/api/data/'

    def __init__(self, api_key: str) -> None:
        self._api_key = api_key
        super().__init__()

    def send_request(self, params: dict) -> dict:
        r"""Base method for sending a GET request to the API.


        This method sends a GET request to the base url with the params + the API key +
        JSON format and handles error messages. If successful, returns the Results
        part of the response.

        Parameters
        ----------
        params
            Additional params to add to the GET request
        """

        params['userid'] = self._api_key
        params['resultformat'] = 'JSON'

        # Send request and get results
        r = requests.get(self._base_url, params=params)

        # Raise in case of unsuccesful status
        r.raise_for_status()

        json_response = r.json()['BEAAPI']

        # Make sure we have results, and no error
        if 'Error' not in json_response:
            results = json_response['Results']
        else:
            results = json_response

        # Check for errors
        if 'Error' in results:
            # Description
            if 'ErrorDetail' in results['Error']:
                descr = results['Error']['APIErrorDescription'] + ' ' +\
                    results['Error']['ErrorDetail']['Description']
            elif 'APIErrorDescription' in results['Error']:
                descr = results['Error']['APIErrorDescription']
            else:
                descr = results['Error']['error']

            # Error number
            if 'APIErrorCode' in results['Error']:
                num = results['Error']['APIErrorCode']
            else:
                num = ''

            raise BEAError(num, descr)

        # ATTN:
        # This checks for the rare case of the IIP and IntlServTrade databases, where for some
        # reason, Data does not appear in Results, but alongside them.

        if 'Data' in json_response:
            results = json_response

        return results


class BEAError(Exception):
    """Error class to handle BEA API specific errors"""

    def __init__(self, status_code: str, message: str) -> None:
        self.status_code = status_code
        self.message = message

        super().__init__()

    def __str__(self):
        return('{0}\n Error code: {1}'.format(
            self.message, self.status_code))
