"""Higher level module to handle data processing of BEA API results"""

import warnings

import pandas as pd

from . import api
from .utilities import *


class BEAResults(object):
    r"""This class holds the results from the
    :meth:`BEA.get_data<beapy.BEA.get_data>` call.
    It implements a special processing for each dataset, see notes bellow.

    Parameters
    ----------
    results
        The dictionary of results for get_data() call obtained from the API
    dataset
        dataset from which the results were obtained
    """

    def __init__(self, results: dict, dataset: str) -> None:
        dataset = dataset.lower()

        # If no data returned, raise error
        if 'Data' not in results:
            raise api.BEAError('', 'No data was returned, you might have set wrong arguments.')

        # Proceed to processing data
        if dataset in ['nipa', 'niunderlyingdetail', 'fixedassets']:
            self._process_NIPA(results)
        elif dataset == 'mne':
            self._process_MNE(results)
        elif dataset == 'gdpbyindustry':
            self._process_GDPbyIndustry(results)
        elif dataset in ['ita', 'iip', 'intlservtrade']:
            self._process_ITA(results)
        elif dataset in ['inputoutput']:
            self._process_InputOutput(results)
        elif dataset in ['regional']:
            self._process_Regional(results)
        else:
            self._just_return(results)

        # Add notes
        if 'Notes' in results:
            self.notes = results['Notes']

    def _just_return(self, results: dict, dataset: str) -> None:
        """A default function, which just attempts to convert the Data part of results to a
        DataFrame. Used if the dataset does not match any of the supported ones, will issue a
        warning to the user.
        """

        self._data = pd.DataFrame(results['Data'])

        warn('''The {} dataset is not supported, a simple conversion of results to a DataFrame
        will be returned'''.format(dataset))

    def _process_main(self, results: dict, index: list, delete: list = []) -> None:
        r"""Common processing, shared by some datasets (NIPA, FA, NIU, ITA, ...)
        What it does is basically:
            - properly format TimePeriod and DataValue
            - take care of DataValue scaling using UNIT_MULT
            - delete some columns
            - set some columns as index, sort by index
        """

        data = pd.DataFrame(results['Data'])

        # Convert some values
        data['TimePeriod'] = pd.to_datetime(data['TimePeriod'])
        data['DataValue'] = pd.to_numeric(data['DataValue'].str.replace(',', ''),
                                          errors='coerce')
        data['UNIT_MULT'] = 10**data['UNIT_MULT'].astype(int)

        #  Get data to base 1
        data['DataValue'] = data['DataValue'] * data['UNIT_MULT']

        # Remove unnecessary columns
        data.drop(columns=['UNIT_MULT'] + delete, errors='ignore', inplace=True)

        # Set index and sort
        data.set_index(index, inplace=True)
        data.sort_index(inplace=True)

        # Set internal parameter for data
        self._data = data

    def _process_NIPA(self, results: dict) -> None:
        """The special processing function for NIPA, NIUnderlyingDetail and FixedAssets
        datasets. It does what _process_main does + deletes the TableName column
        (as multiple tables are not allowed, it is not needed). For index it uses TimePeriod,
        LineDescription and LineNumber. It also convers LineNumber to numeric."""

        self._process_main(results, delete=['TableName'],
                           index=['TimePeriod', 'LineNumber', 'LineDescription'])

        da = self._data
        da.reset_index(level=['LineDescription', 'LineNumber'], inplace=True)
        da['LineNumber'] = pd.to_numeric(da['LineNumber'])
        da.set_index(['LineNumber', 'LineDescription'], append=True, inplace=True)
        da.sort_index(inplace=True)

    def _process_MNE(self, results: dict) -> None:
        r"""Specific function for the MNE dataset.
        Does the following:
            - Splits the DataValue column to ones with actual values and ones with Footnotes
            - Takes care of scaling: tries to infer the scale (and adjust DataValue)
              based on TableScale column, issues warning if this is not possible for all items
            - Sets index based on Year (TimePeriod), Row, Column and SeriesName
            - Deletes/renames/transforms type for some columns
            - Sort by index columns

        CANT GET OLD SIC CLASSIFICATION (TEST 2)
        """

        data = pd.DataFrame(results['Data'])
        index = ['TimePeriod', 'Row', 'Column', 'SeriesName']

        # Drop unnecessary columns, rename
        data.drop(columns=['TableColumnDisplayOrder', 'TableRowDisplayOrder', 'DataValue'],
                  inplace=True)
        data.rename(columns={'DataValueUnformatted': 'DataValue', 'Year': 'TimePeriod'},
                    inplace=True)

        # Separate pure numbers from footnotes
        only_number = data['DataValue'].str.contains(r'^-?\d*\.?\d*$')
        data['Footnotes'] = data.loc[~only_number, 'DataValue']
        data.loc[~only_number, 'DataValue'] = None

        # Transform a few values to numeric/date
        data['SeriesID'] = data['SeriesID'].astype(int)
        data['TimePeriod'] = pd.to_datetime(data['TimePeriod'])
        data['DataValue'] = pd.to_numeric(data['DataValue'], errors='coerce')

        # Take care of scaling, if possible -- otherwise issue a warning
        scales = detect_scales(data['TableScale'])

        if scales['scale'].isna().any() or scales['unit'].isna().any():
            msg = 'Could not determine the scale/unit of all entries, returning scaled data.'
            warnings.warn(msg)
        else:
            data['DataValue'] = data['DataValue'] * (10**scales['scale'])
            data['TableScale'] = scales['unit']

        # Drop (grand) parent columns, if possible
        if (data['ColumnGParent'].unique() == ['None']).all():
            data.drop(columns=['ColumnGParent', 'ColumnGParentCode'], inplace=True)
        else:
            index += ['ColumnGParent']

        if (data['ColumnParent'].unique() == ['None']).all():
            data.drop(columns=['ColumnParent', 'ColumnParentCode'], inplace=True)
        else:
            index += ['ColumnParent']

        # Set index, sort
        data.set_index(index, inplace=True)
        data.sort_index(inplace=True)

        # Set parameter
        self._data = data

    def _process_GDPbyIndustry(self, results: dict) -> None:
        r"""Specific function for the GDPbyIndustry dataset.
        Does the following:
            - Properly formats the date (either year or quarter)
            - Transforms DataValue to numeric
            - Sets index based on TableID, TimePeriod and IndustrYDescription
            - Sort by index columns
        """

        data = pd.DataFrame(results['Data'])

        # Format date properly
        if data.loc[0, 'Frequency'] == 'A':
            data['TimePeriod'] = pd.to_datetime(data['Year'])
        elif data.loc[0, 'Frequency'] == 'Q':
            roman_dict = {'I': '1', 'II': '2', 'III': '3', 'IV': '4'}
            data['TimePeriod'] = pd.to_datetime(
                data['Year'] + 'Q' + data['Quarter'].map(roman_dict))
        else:
            raise TypeError('Unrecognized frequency, only Q and A are supported.')

        data.drop(columns=['Year', 'Quarter', 'Frequency'], inplace=True)

        # Format DataValue
        data['DataValue'] = pd.to_numeric(data['DataValue'], errors='coerce')

        # Set index, sort
        data.set_index(['TableID', 'TimePeriod', 'IndustrYDescription'], inplace=True)
        data.sort_index(inplace=True)

        # Warn about scale
        warnings.warn('Scale of data can not be inferred for the GDPbyIndustry dataset. ' +
                      'Check footnotes.')

        # Set parameter
        self._data = data

    def _process_ITA(self, results: dict) -> None:
        """Specific function for the ITA, IIP and IntlServTrade datasets.
        It uses the _process_main function, setting as index TimePeriod and TimeSeriesId,
        and deleting Year and Frequency
        """

        self._process_main(results, delete=['Year', 'Frequency'],
                           index=['TimePeriod', 'TimeSeriesId'])

    def _process_InputOutput(self, results: dict) -> None:
        """Specific function for the InputOutput dataset.
        It does the following:
        - Formats the year as a date, DataValue as numeric
        - Replaces the name of 'ColDescr' and 'RowDescr' with the name in 'ColType' and 'RowType'
          respectively. Does the same (but adds 'Code') for 'ColCode' and 'RowCode'
        - Sets the TableID, Year, column and row as index
        - deletes Year, Rowtype and Coltype (if possible)

        Unfortunately, no information about the scale of the data is provided, so a warning
        is issued to the user.
        """

        data = pd.DataFrame(results['Data'])
        drop = ['Year']

        # Format columns
        data['TimePeriod'] = pd.to_datetime(data['Year'])
        data['DataValue'] = pd.to_numeric(data['DataValue'].str.replace(',', ''),
                                          errors='coerce')

        # Replace column/row names
        col_index = 'ColDescr'
        row_index = 'RowDescr'

        if len(data['RowType'].unique()) == 1:
            row_index = data['RowType'][0]
            data.rename(columns={'RowCode': row_index + 'Code', 'RowDescr': row_index},
                        inplace=True)
            drop += ['RowType']

        if len(data['ColType'].unique()) == 1:
            col_index = data['ColType'][0]

            if col_index == row_index:
                col_index = row_index + '_1'

            data.rename(columns={'ColCode': col_index + 'Code', 'ColDescr': col_index},
                        inplace=True)
            drop += ['ColType']

        # Set and sort index
        data.set_index(['TableID', 'TimePeriod', col_index, row_index], inplace=True)
        data.sort_index(inplace=True)

        # Delete columns
        data.drop(columns=drop, errors='ignore', inplace=True)

        # Warn about scale
        msg = 'Scale of data can not be inferred for the InputOutput dataset. Check footnotes'
        warnings.warn(msg)

        # Set internal parameter for data
        self._data = data

    def _process_Regional(self, results: dict) -> None:
        """Specific function for the Regional dataset. It uses _process_main to set TimePeriod,
        Code and GeoName as index. An attempt is made to remove the scale from CL_UNIT.
        """

        self._process_main(results, ['TimePeriod', 'Code', 'GeoName'])

    def data(self, with_details: bool = False) -> pd.DataFrame:
        r"""This method returns the DataFrame containing the data. It has an optional argument
        ``with_details``, which determines whether non-data columns are returned as well.

        Parameters
        ----------
        with_details: False
            Whether or not to return non-data columns

        Returns
        -------
        pd.DataFrame
            A DataFrame with a description and name of availible parameters,
            as well as some other options.
        """

        if not with_details:
            return self._data[['DataValue']]
        else:
            return self._data


class BEA(api.BEAAPI):
    r"""The primary class in the BEApy package, handles sending requests
    and manipulating the results. Needs to be initiated with your BEA
    API key once, it is stored for all later requests.

    Parameters
    ----------
    api_key
        Your BEA API key
    """

    def __init__(self, api_key: str) -> None:
        super().__init__(api_key)

    def dataset_list(self) -> pd.DataFrame:
        r"""Get a list of all availible datasets (API method getDatasetList)

        Returns
        -------
        pd.DataFrame
            A DataFrame with a description and name of availible datasets.
        """

        params = {
            'method': 'GETDATASETLIST'
        }

        results = self.send_request(params)['Dataset']

        return json_result_to_dataframe(results)

    def parameter_list(self, dataset: str) -> pd.DataFrame:
        r"""Get a list of availible parameters for the dataset  (API method getParameterList)

        Usually the parameter list will also include the value of the ALL option,
        as well as data type, whether the parameteris is required, and
        whether it accepts multiple values.

        Parameters
        ----------
        dataset
            The name of the dataset

        Returns
        -------
        pd.DataFrame
            A DataFrame with a description and name of availible parameters,
            as well as some other options.
        """

        params = {
            'method': 'GETPARAMETERLIST',
            'datasetName': dataset,
        }

        results = self.send_request(params)['Parameter']

        parameters = json_result_to_dataframe(results)

        # Shorten names
        parameters.columns = [x.replace('Parameter', '') for x in parameters.columns]
        parameters.columns = [x.replace('Flag', '') for x in parameters.columns]
        parameters['Name'] = parameters['Name'].str.strip()

        # Set name as index
        parameters.set_index('Name', inplace=True)

        # Default column order, reorder according to it
        default_order = ['Description', 'DefaultValue', 'DataType',
                         'IsRequired', 'MultipleAccepted', 'AllValue']
        parameters = reorder_columns(parameters, default_order)

        # Columns to bool
        parameters['MultipleAccepted'] = parameters['MultipleAccepted']\
            .astype(int).astype(bool)
        parameters['IsRequired'] = parameters['IsRequired']\
            .astype(int).astype(bool)

        return parameters

    def parameter_values(self, dataset: str, parameter: str, **kwargs) -> pd.DataFrame:
        r"""Get a list of possible values for the parameter in a dataset (API method
        getParameterValues).

        Optionally, filter the list based on values of other parameters (API method
        getParameterValuesFiltered). Be careful in this case, as the filtered method is not
        implemented for all datasets.

        Parameters
        ----------
        dataset
            The name of the dataset
        parameter
            The name of the (target) parameter whose values you are retrieving
        kwargs
            Additional arguments (parameters), based on which to filter the target parameter. All
            arguments should be strings or lists of strings. If ``kwags`` is empty,
            getParameterValues API method is used. Otherwise getParameterValuesFiltered
            is used.

        Returns
        -------
        pd.DataFrame
            A DataFrame with a description and name (key) of availible parameters values
        """

        # Set up things differently, depending on whether we are filtering or not
        if len(kwargs) == 0:
            method = 'GETPARAMETERVALUES'
            param_key = 'parameterName'
        else:
            # Make sure all the parameter values are strings, or lists of strings
            if not check_all_types(kwargs.values(), str):
                raise TypeError('All parameter values must be strings or lists/tuples of strings')

            method = 'GETPARAMETERVALUESFILTERED'
            param_key = 'targetParameter'

        params = {
            'method': method,
            'datasetName': dataset,
            param_key: parameter
        }

        if kwargs is not None:
            params.update(kwargs)

        results = self.send_request(params)['ParamValue']

        values = json_result_to_dataframe(results)

        return values

    def find_table(self, dataset: str, search: str) -> pd.DataFrame:
        r"""A helper function to search for a table based on its description (or some words
        from it).

        This function will return all tables from a dataset, whose description (as would be
        returned by :meth:`parameter_values<BEA.parameter_values>`) contains the ``search`` string. This can be useful to
        find a table based on its name on the BEA website.

        The supported datasets (ones with tables) are: NIPA, NIUnderlyingDetail, FixedAssets,
        GDPbyIndustry, UnderlyingGDPbyIndustry, InputOutput and Regional.

        Additionaly, the function also performs the search on some table-like parameters of other
        datasets (Indicator for ITA, TypeOfInvestment for IIP, TypeOfService for
        IntlServTrade and SeriesID for MNE).

        Parameters
        ----------
        dataset
            The name of the dataset
        parameter
            A string which will try to be matched to a substring of tables' description.

        Returns
        -------
        pd.DataFrame
            A DataFrame with a name and description of matching tables
        """

        table_param_name = {
            'NIPA': 'TableName',
            'NIUnderlyingDetail': 'TableName',
            'FixedAssets': 'TableName',
            'GDPbyIndustry': 'TableID',
            'UnderlyingGDPbyIndustry': 'TableID',
            'InputOutput': 'TableID',
            'Regional': 'TableName',
            'ITA': 'Indicator',
            'IIP': 'TypeOfInvestment',
            'IntlServTrade': 'TypeOfService',
            'MNE': 'SeriesID',
        }

        # Transform all to lower case
        dataset = dataset.lower()
        table_param_name = dict_lower(table_param_name)

        if dataset not in table_param_name.keys():
            raise ValueError('find_table does not support the {} dataset'.format(dataset))

        tables = self.parameter_values(dataset=dataset, parameter=table_param_name[dataset])
        descr_col = [col for col in tables.columns if 'desc' in col.lower()][0]

        return tables[tables[descr_col].str.contains(search, case=False)]

    def get_data(self, dataset: str, **kwargs) -> BEAResults:
        r"""Get the data (API method getData).

        Parameters
        ----------
        dataset
            The name of the dataset
        kwargs
            Additional parameters (years, table name, etc.) passed to the API. The values of the
            arguments have to be strings or lists of strings. Additionaly, for datasets that admit
            the Frequency parameter, its value must be set and can not contain multiple values
            (or ALL).

        Returns
        -------
        BEAResults
            A :class:`BEAResults<beapy.BEAResults>` object containing the results.
        """

        params = {
            'method': 'GETDATA',
            'datasetName': dataset
        }

        if not check_all_types(kwargs.values(), str):
            raise TypeError('All parameter values must be strings or lists/tuples of strings')

        params.update(kwargs)

        params = dict_lower(params)

        # Check for multiple frequency arguments, return error
        # Also, require frequency where it is availible
        if 'frequency' in params:
            if isinstance(params['frequency'], list) or ',' in params['frequency']:
                raise ValueError('The Frequency argument should not contain multiple values.' +
                                 'Try splitting the call into one call per each frequency value.')
        else:
            if dataset in ['nipa', 'niunderlyingdetail', 'gdpbyindustry', 'ita', 'iip',
                           'underlyinggdpbyindustry']:
                raise ValueError('Please provide a value for the Frequency parameter.')

        results = self.send_request(params)

        return BEAResults(results, dataset)
