import re

import pandas as pd
import numpy as np


def json_result_to_dataframe(json_data) -> pd.DataFrame:
    """This function converts a list (which we got from JSON) to a DataFrame,
    and takes care of the special case where list had 1 element, so a dict was returned
    """

    if isinstance(json_data, dict):
        df = pd.DataFrame(json_data, index=[0])
    else:
        df = pd.DataFrame(json_data)

    return df


def dict_lower(dictionary: dict, keys: bool = True, values: bool = True) -> dict:
    """For a dictionary with string keys and/or values, this returns a dictionary where all the
    keys and/or values are the lower case version of original keys/values"""

    new_dict = {}
    for key, val in dictionary.items():
        if keys:
            key = key.lower()
        if values and isinstance(values, str):
            val = val.lower()

        new_dict[key] = val

    return new_dict


def check_all_types(items, correct_type):
    """This function checks that all items are of the correct type,
    if not returns False, otherwise True. If item is an iterable (tuple or list),
    it checks the items of that iterable"""

    type_ok = []

    for x in items:
        if isinstance(x, list) or isinstance(x, tuple):
            type_ok.append(all([isinstance(y, str) for y in x]))
        else:
            type_ok.append(isinstance(x, str))

    return all(type_ok)


def detect_scales(series: pd.Series, separator: str = r' of ') -> pd.Series:
    """This function detects if there are some scales (thousands, millions, billions, trillions)
    in the data (as a word, not a nubmer), and returns a series of corresponding exponents
    (3, 6, 9 or 10) and the units. The expected format of the data is something like
    "Millions of Dollars"."""

    units_dict = {'thousands': 3, 'millions': 6, 'billions': 9, 'trillions': 12}

    regex_str = r'^(?P<scale>' + r'|'.join(units_dict.keys()) + ')' + separator + r'(?P<unit>.+)$'
    results = series.str.extract(regex_str, flags=re.IGNORECASE)

    results['scale'] = results['scale'].str.lower().map(units_dict)

    return results


def reorder_columns(df: pd.DataFrame, order: list) -> pd.DataFrame:
    """This function reorders the columns of the dataframe by the order specified in the ``order``
    argument. The columns not in ``order`` are moved after the ones in ``order`` (but retain
    their relative ordering). If some column names in ``order`` are not present in ``df``, no
    errors are raised - these columns are simply ignored."""

    # Default column order - filter columns not availible, append others
    new_order = [x for x in order if x in df.columns]
    extra = [x for x in df.columns if x not in order]

    new_order = new_order + extra

    # Reorder and return
    return df[new_order]
