'''This file tests the data-related functions of the package.

First a few cases of catching errors are tested, then, for all the supported datasets,
a test class is written that makes sure that the output for the NIPA dataset matches the expected
(real) one, which was confirmed with 'manual' inspection for a few tables.

The data tested is the data from the earlies date availible until 2018.
'''

import pytest
import pandas as pd

from beapy import BEA
from beapy.api import BEAError


def assert_without_col_order(right, left):
    """This function will make sure that column order does not affect the
    assert statement for two pandas DataFrames"""

    if set(right.columns) == set(left.columns):
        right = right[left.columns]

    pd.testing.assert_frame_equal(right, left, check_dtype=False)

# This gets all the years up to the one before the current one
@pytest.fixture(scope='module')
def years():
    return [str(x) for x in range(1900, 2019)]


class TestWrongGetData(object):
    """This class tries to catch the various errors that should appear
    in some calls to get_data(), such as
    - Wrong or missing frequency argument
    - Empty data being returned
    - Non-string arguments
    - Missing required arguments
    """

    def test_missing_frequency(self, bea):
        with pytest.raises(BEAError):
            bea.get_data(dataset='NIPA', year='X')

    def test_multiple_frequencies(self, bea):
        with pytest.raises(ValueError):
            bea.get_data(dataset='NIPA', year='X', frequency=['A', 'Q'])

    def test_wrong_type_single_arg_type(self, bea):
        with pytest.raises(TypeError):
            bea.get_data(dataset='Regional', year=2017)

    def test_wrong_type_list_arg_type(self, bea):
        with pytest.raises(TypeError):
            bea.get_data(dataset='Regional', year=['2016', 2017])

    def test_missing_argument(self, bea):
        with pytest.raises(BEAError):
            bea.get_data(dataset='Regional', year='2016', GeoFips='00000')


class TestNIPAData(object):

    def test_T10103_Q(self, bea, years):
        get_data = bea.get_data(dataset='NIPA', tableName='T10103', frequency='Q', year=years)
        true_data = pd.read_hdf('tests/Datasets/NIPA/T10103_Q.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))

    def test_T20505_A(self, bea, years):
        get_data = bea.get_data(dataset='NIPA', tableName='T20505', frequency='A', year=years)
        true_data = pd.read_hdf('tests/Datasets/NIPA/T20505_A.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))

    def test_T50303_Q(self, bea, years):
        get_data = bea.get_data(dataset='NIPA', tableName='T50303', frequency='Q', year=years)
        true_data = pd.read_hdf('tests/Datasets/NIPA/T50303_Q.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))


class TestFA(object):

    def test_FAAt105_A(self, bea, years):
        get_data = bea.get_data(dataset='FixedAssets', tableName='FAAt105', frequency='A',
                                year=years)
        true_data = pd.read_hdf('tests/Datasets/FixedAssets/FAAt105_A.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))

    def test_FAAt309ESI_A_A(self, bea, years):
        get_data = bea.get_data(dataset='FixedAssets', tableName='FAAt309ESI', frequency='A',
                                year=years)
        true_data = pd.read_hdf('tests/Datasets/FixedAssets/FAAt309ESI_A.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))


class TestMNE(object):

    def test_MNE_EMP_OUT_CbI(self, bea, years):
        get_data = bea.get_data(dataset='MNE', DirectionOfInvestment='inward',
                                classification='CountryByIndustry', SeriesID='8',
                                Country='all', NonbankAffiliatesOnly='0', ownershiplevel='0',
                                year=years)
        true_data = pd.read_hdf('tests/Datasets/MNE/MNE_EMP_IN_CbI.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))

    def test_MNE_ReInv_OUT_CbI(self, bea, years):
        get_data = bea.get_data(dataset='MNE', DirectionOfInvestment='outward',
                                classification='CountryByIndustry', SeriesID='25', Country='all',
                                year=years)
        true_data = pd.read_hdf('tests/Datasets/MNE/MNE_ReInv_OUT_CbI.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))

    def test_MNE_EMPNew_ReInv_IN_SbT(self, bea, years):
        get_data = bea.get_data(dataset='MNE', DirectionOfInvestment='inward',
                                classification='StateByType', SeriesID='64', Country='all',
                                year=years)
        true_data = pd.read_hdf('tests/Datasets/MNE/MNE_EMPNew_IN_SbT.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))

    def test_combined(self, bea, years):
        """This test basically makes sure that combining two series with different units still
        produces sensible scales"""

        get_data = bea.get_data(dataset='MNE', DirectionOfInvestment='inward',
                                classification='IndustrybyCountryofUBO', SeriesID=['1', '8'],
                                Country='all', NonbankAffiliatesOnly='0', ownershiplevel='0',
                                year=years)
        true_data = pd.read_hdf('tests/Datasets/MNE/MNE_combined.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))


class TestGDPByIndustry(object):

    def test_VA_Q(self, bea, years):
        get_data = bea.get_data(dataset='GDPByIndustry', TableID='1', Frequency='Q',
                                Industry='ALL', year=years)
        true_data = pd.read_hdf('tests/Datasets/GDPByIndustry/VA_Q.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))

    def test_GObI_A(self, bea, years):
        get_data = bea.get_data(dataset='GDPByIndustry', TableID='16', Frequency='A',
                                Industry='ALL', year=years)
        true_data = pd.read_hdf('tests/Datasets/GDPByIndustry/GObI_A.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))


class TestITA(object):

    def test_ExpGdsServIncRec_QSA(self, bea, years):
        get_data = bea.get_data(dataset='ita', Frequency='QSA', Indicator=['ExpGdsServIncRec'],
                                year=years)
        true_data = pd.read_hdf('tests/Datasets/ITA/ExpGdsServIncRec_QSA.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))

    def test_ExpServInsurance_A(self, bea, years):
        get_data = bea.get_data(dataset='ita', Frequency='A', Indicator=['ExpServInsurance'],
                                year=years)
        true_data = pd.read_hdf('tests/Datasets/ITA/ExpServInsurance_A.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))

    def test_ExpServInsurance_A(self, bea, years):
        get_data = bea.get_data(dataset='ita', Frequency='QNSA',
                                Indicator=['DiInvReinvestEarnAssets'], year=years)
        true_data = pd.read_hdf('tests/Datasets/ITA/DiInvReinvestEarnAssets_QNSA.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))


class TestIIP(object):

    def test_PfInvAssets_QNSA(self, bea, years):
        get_data = bea.get_data(dataset='iip', Frequency='QNSA', TypeOfInvestment=['PfInvAssets'],
                                Component='Pos', year=years)
        true_data = pd.read_hdf('tests/Datasets/IIP/PfInvAssets_QNSA.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))

    def test_PfInvAssets_QNSA(self, bea, years):
        get_data = bea.get_data(dataset='iip', Frequency='A', TypeOfInvestment=['FinLiabsFoa'],
                                Component='Pos', year=years)
        true_data = pd.read_hdf('tests/Datasets/IIP/FinLiabsFoa_A.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))


class TestIntlServTrade(object):

    def test_ictserv(self, bea, years):
        get_data = bea.get_data(dataset='IntlServTrade', TypeOfService='IctServ', year=years)
        true_data = pd.read_hdf('tests/Datasets/IntlServTrade/IctServ.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))

    def test_mining(self, bea, years):
        get_data = bea.get_data(dataset='IntlServTrade', TypeOfService='Mining', year=years)
        true_data = pd.read_hdf('tests/Datasets/IntlServTrade/Mining.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))


class TestInputOutput(object):

    def test_use_commodities(self, bea, years):
        get_data = bea.get_data(dataset='InputOutput', TableID='258', year=years)
        true_data = pd.read_hdf('tests/Datasets/InputOutput/258.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))

    def test_ind_requirements(self, bea, years):
        get_data = bea.get_data(dataset='InputOutput', TableID='60', year=years)
        true_data = pd.read_hdf('tests/Datasets/InputOutput/60.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))


class TestRegional(object):
    # Include an example of how to get all counties in alabama
    def test_SQGDP2_1(self, bea, years):
        get_data = bea.get_data(dataset='Regional', TableName='SQGDP2', GeoFips='STATE',
                                LineCode='1', year=years)
        true_data = pd.read_hdf('tests/Datasets/Regional/SQGDP2_1.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))

    def test_MAGDP8_3(self, bea, years):
        get_data = bea.get_data(dataset='Regional', TableName='MAGDP8', GeoFips='MSA',
                                LineCode='3', year=years)
        true_data = pd.read_hdf('tests/Datasets/Regional/MAGDP8_3.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))

    def test_CAEMP25N_10(self, bea, years):
        get_data = bea.get_data(dataset='Regional', TableName='CAEMP25N', GeoFips='COUNTY',
                                LineCode='10', year=years)
        true_data = pd.read_hdf('tests/Datasets/Regional/CAEMP25N_10.h5')

        assert_without_col_order(true_data, get_data.data(with_details=True))
